/* dumpmem - a small utility to execute a binary and dump its memory */ 
/* contents before it begins executing.                              */ 
/*                                                                   */ 
/* Muli Ben-Yehuda <mulix@mulix.org>                                 */ 
/* Licensed under the GPL V2                                         */ 
/*                                                                   */ 
/* $Id: dumpmem.c,v 1.10 2002/11/16 10:04:01 mulix Exp $              */ 

#include <stdio.h> 
#include <string.h> 
#include <fcntl.h>
#include <stdlib.h> 
#include <unistd.h> 
#include <errno.h> 
#include <assert.h> 
#include <sys/stat.h>
#include <sys/ptrace.h> 
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h> 
#include <sys/syscall.h>    /* For SYS_write etc */
#include <linux/user.h>     /* For constants ORIG_EAX etc */
#define _GNU_SOURCE
#include <getopt.h>

#include "util.h" 
#include "maps.h" 

const int word_size = sizeof(long); 

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
#endif /* ARRAY_SIZE */ 

struct config 
{
	int raw; /* 0 on, 1 off */ 
	char* output_fname; 
	FILE* output_fd; 
	int attach; /* attach or exec? */
	union { 
		char** child_args; /* program to exec and its cmdline args */ 
		pid_t pid; /* process to attach to */ 
	} target; 
}; 

/* make a ptrace request and barf if it fails */ 
static unsigned long xptrace(enum __ptrace_request req, pid_t proc, void* addr)
{
	unsigned long res; 

	/* global errno */ 
	errno = 0; 
	res = ptrace(req, proc, addr, NULL); 
	if (res == (unsigned long)-1 && errno)
		fatal("ptrace(proc %d, req %d): %s", proc, req, strerror(errno)); 

	return res; 
}

/* get the data at 'addr */ 
static unsigned long getmem(pid_t proc, unsigned long addr)
{
	return xptrace(PTRACE_PEEKDATA, proc, (void*)addr); 
}

/* dump the vma described by this map */ 
static void dump_formatted_vma(pid_t proc, struct map* m, FILE* out)
{
	unsigned long addr = m->begin; 
	int counter = 0; 
	fprintf(out, "%08lx: ", addr); 
	do {
		unsigned long dat; 
		if (++counter == 7) {
			fprintf(out, "\n%08lx: ", addr); 
			counter = 0; 
		}

		dat = getmem(proc, addr); 

		addr = addr + word_size; 

		fprintf(out, "%08lx ", dat); 

	} while (addr < m->end); 

	fprintf(out, "\n"); 
}

static void dump_raw_vma(pid_t proc, struct map* m, FILE* out)
{
	unsigned long addr = m->begin; 
	unsigned long long begin, end, span; 

	begin = m->begin; 
	end = m->end; 
	span = m->end - m->begin; 

	/* output the vma header */ 
	fwrite(&begin, sizeof(begin), 1, out); 
	fwrite(&end, sizeof(end), 1, out); 
	fwrite(&span, sizeof(span), 1, out); 

	/* output this vma */ 
	do {
		unsigned long dat = getmem(proc, addr); 

		fwrite(&dat, sizeof(dat), 1, out); 
		addr = addr + word_size; 
		span -= word_size; 

	} while (addr < m->end); 
}

static void dump_vma(pid_t proc, struct map* m, struct config* cfg)
{
	if (cfg->raw)
		dump_raw_vma(proc, m, cfg->output_fd); 
	else 
		dump_formatted_vma(proc, m, cfg->output_fd); 
}

/* dump the entire memory for this process */ 
static void dump_proc_mem(pid_t proc, struct config* cfg)
{
	struct map *p, *del; 
	get_maps(proc, &p); 

	if (!cfg->output_fname)
		cfg->output_fd = stdout; 
	else {
		if (!(cfg->output_fd = fopen(cfg->output_fname, "w")))
			fatal("could not open output file '%s' for writing", 
			      cfg->output_fname); 
	}
		
	while (p) {
		if (!cfg->raw) 
			fprintf(cfg->output_fd, "map: 0x%08lx-0x%08lx\n\n", 
				p->begin, p->end); 
		
		dump_vma(proc, p, cfg); 

		del = p; 
		p = p->next; 
		
		free(del); 

		if (!cfg->raw)
			fprintf(cfg->output_fd, "\n"); 
	}

	fflush(cfg->output_fd); 

	if (cfg->output_fname)
		fclose(cfg->output_fd); 
}

/* copied from syscalltrack's sctrace/main.cpp. I can do that, I wrote it */ 
static char* find_in_path(const char* name)
{
	char* ret = NULL;
	char* env = getenv("PATH"); 
	char* beg, *end; 
	char buf[MAXPATHLEN + 256]; 

	if (!env)
		return NULL; 

	if (!(env = strdup(env)))
		return NULL; 

	beg = env; 

	/* find the first : character. */
	while (1) {
		end = strchr(beg, ':'); 
		if (end)
			*end = '\0'; 

		(void)snprintf(buf, sizeof(buf), "%s/%s", beg, name); 

		if (!access(buf, X_OK)) {
			ret = strdup(buf); 
			break; 
		}
		if (!end)
			break; 
		++end; 
		beg = end; 
	}
	free(env); 
	return ret; 
}

void exec_child(const char* exec, char** argv)
{
	const char* fullpath; 

	xptrace(PTRACE_TRACEME, 0, NULL); 

	if (*exec == '/')
		fullpath = exec; 
	else if ((*exec == '.') && 
		 ((strlen(exec) > 1) && (exec[1] == '.' || exec[1] == '/')))
		fullpath = exec; 
	else if (!(fullpath = find_in_path(exec)))
		fatal("could not find file to execute '%s'\n", exec); 

	printf("starting"); 
	execv(fullpath, argv); 
	fatal("execv %s: %s", fullpath, strerror(errno)); 
}

static void run_tracer(pid_t child, struct config* cfg)
{
	int status; 
	unsigned long orig_eax; 
	int insyscall;
	int res; 

	res = waitpid(child, &status, 0);
	if (res != child)
		fatal("waitpid"); 
	
	if (WIFEXITED(status))
		return; 

	orig_eax = xptrace(PTRACE_PEEKUSER, child, (void*)(4 * ORIG_EAX)); 
	if (orig_eax == SYS_execve) {
		if(!insyscall) {    /* syscall entry */
			insyscall = 1;
			xptrace(PTRACE_SYSCALL, child, NULL); 
		} else { /* syscall exit, maps set, do our thing */ 
			dump_proc_mem(child, cfg); 
			xptrace(PTRACE_KILL, child, NULL); 
		}
	}
	
	do {
		res = waitpid(child, &status, 0); 
		if (res != child)
			fatal("waitpid: %s", strerror(errno)); 
	} while (!WIFEXITED(status) && !WIFSIGNALED(status)); 
}

static void usage(const char* prog)
{
	static char usagestr[] = 
		"%s (%s) is process memory dumper\n"
		"Usage: %s [OPTIONS] [program_to_execute]\n"
		"\n"
		"Options: \n"
		"  -o, --output file_name      send the output to this file\n"
                "  -r, --raw                   send the output in raw format\n"
		"  -p, --process pid           pid of the process we\n"
		"                              should attach to\n" 
		"  -h, --help                  this help message\n"; 
	
	fprintf(stdout, usagestr, prog, VERSION_STR, prog);
	exit(EXIT_FAILURE); 
}

static void parse_cmdline(int argc, char* argv[], struct config* cfg)
{
	unsigned long p; 
	const char* progname = argv[0]; 

	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{"output", 1, NULL, 'o'}, 
			{"raw", 0, NULL, 'r'},
			{"process", 1, NULL, 'p'}, 
			{"help", 0, NULL, 'h'},
			{NULL, 0, NULL, 0}
		};
		
		/* see getopt(3) for why '+': in short, it tells getopt to stop 
		   as soon as the first non option argument is found */ 
		c = getopt_long (argc, argv, "+o:rp:h",
				 long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c) {
		case 0:
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
			
		case 'r':
			cfg->raw = 1; 
			break;
		case 'o':
			if (!(cfg->output_fname = strdup(optarg)))
				fatal("oom"); 
			break; 
		case 'p': 
			/* enforce 16 bit pids */ 
			if ((p  = strtol(optarg, NULL, 0)) & ~0xffff) {
				fprintf(stdout, "sorry, pid %ld does not appear "
					"to be a valid (16 bit) pid\n\n", p); 
				usage(progname); 
			}
			cfg->attach = 1; 
			cfg->target.pid = (pid_t)p; 
			break; 
		case 'h':
			usage(progname); 
			break; 
		case '?':
			usage(progname); 
			break;
		default:
			printf ("?? getopt returned character code 0%o ??\n", c);
		}
	}

	if (!cfg->attach) {
		/* the next argument is the name of the prog to exec */ 
		if (optind >= argc) {
			fprintf(stdout, "ahm, no program to execute given\n\n"); 
			usage(progname);
		} else 
			cfg->target.child_args = &argv[optind]; 
	}
}

void attach_to_proc(struct config* cfg)
{
	unsigned long pret; 
	pid_t pid = cfg->target.pid; 

	/* check that this pid exists */ 
	if (kill(pid, 0))
		fatal("could not atttach to pid %d", pid); 

	/* attach to the running process */ 
	/* yes, we have a race here, but it's harmless */ 
	if ((pret = xptrace(PTRACE_ATTACH, pid, NULL)))
		fatal("could not attach to pid %d", pid); 

#if 0 /* I don't know if I need it or not... */ 
/* I don't really care if it's stopped or not */
	res = waitpid(pid, &status, 0);
	if (res != pid)
		fatal("waitpid"); 
#endif /* 0 */

	/* dump its memory */ 
	dump_proc_mem(pid, cfg); 

	/* tell it to continue */ 
	if ((pret = xptrace(PTRACE_DETACH, pid, NULL)))
		fatal("could not detach from pid %d", pid); 
}

void trace_new_proc(struct config* cfg)
{
	pid_t child = fork(); 
	if (child == -1)
		fatal("fork failed\n"); 
	
	if (child == 0) {
		/* child code */ 
		char** args = cfg->target.child_args; 
		exec_child(*args, args); 
	} else 
		run_tracer(child, cfg); 
}

/* main */ 
int main(int argc, char* argv[])
{
	struct config cfg; 

	memset(&cfg, 0, sizeof(cfg)); 
	parse_cmdline(argc, argv, &cfg); 

	if (cfg.attach)
		attach_to_proc(&cfg); 
	else 
		trace_new_proc(&cfg); 

	return 0; 
}
